# Generated by Django 3.2.13 on 2022-05-23 04:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("document_templates", "0003_auto_20220328_0930"),
    ]

    operations = [
        migrations.AddField(
            model_name="documenttemplatevariable",
            name="template",
            field=models.TextField(
                blank=True,
                help_text="Jinja template with 'value' in context",
                verbose_name="Template",
            ),
        ),
    ]



#!/usr/bin/env python
"""
To regenerate that module to contain the latest list of  IANA Language
Subtag Registry, either call this module directly from the command line
(``python regenenerate.py``), or call the ``regenerate`` method.

"""
import os
import re
import urllib
import urllib.request

TEMPLATE = """from django.utils.translation import gettext as _

LANGUAGES = (
    %(languages)s
)

"""


def regenerate():
    """
    Generate the languages Python module.
    """
    paren = re.compile(r"\([^)]*\)")
    location = "http://www.iana.org/assignments/language-subtag-registry"

    # Get the language list.
    with urllib.request.urlopen(location) as f:  # nosec # because hardcoded
        lines = f.read().decode().splitlines()
    languages = []
    info = {}
    p = None
    for line in lines:
        if line == "%%":
            if (
                "Type" in info
                and info["Type"] == "language"
                and info["Description"] != "Private use"
            ):
                languages.append(info)
            info = {}
        elif ":" not in line and p:
            info[p[0]] = paren.sub("", p[2] + line).strip()
        else:
            p = line.partition(":")
            if (
                not p[0] in info
            ):  # Keep the first description as it should be the most common
                info[p[0]] = paren.sub("", p[2]).strip()

    languages_lines = map(
        lambda x: '("%s", _(u"%s")),' % (x["Subtag"], x["Description"]), languages
    )

    # Generate and save the file.
    filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), "languages.py")
    # TODO: first make a backup of the file if it exists already.
    with open(filename, "w") as f:
        f.write(
            TEMPLATE
            % {
                "languages": "\n    ".join(languages_lines),
            }
        )


if __name__ == "__main__":
    regenerate()



# Generated by Django 3.2.12 on 2022-03-28 03:33

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('reports', '0005_auto_20210826_1951'),
    ]

    operations = [
        migrations.AlterField(
            model_name='report',
            name='model',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='contenttypes.contenttype', verbose_name='Model'),
        ),
    ]



import datetime
import numbers
import os
from collections.abc import Iterable

import htmlgenerator as hg
from django.conf import settings
from django.db import models
from django.utils.functional import Promise
from django.utils.html import linebreaks, mark_safe
from django.utils.translation import gettext_lazy as _

import bread.settings as app_settings

from . import layout


def format_value(value):
    """
    Renders a python value in a nice way in HTML. If a field-definition has an
    attribute "renderer" set, that function will be used to render the value
    """
    if isinstance(value, bool) or value is None:
        return CONSTANTS[value]

    if isinstance(value, models.Manager):
        value = value.all()

    if isinstance(value, bool) or value is None:
        return CONSTANTS[value]
    if isinstance(value, datetime.timedelta):
        return as_duration(value)
    if isinstance(value, numbers.Number):
        return f"{float(value):f}".rstrip("0").rstrip(".")
    if isinstance(value, models.fields.files.FieldFile):
        return as_download(value)
    if isinstance(value, Iterable) and not isinstance(value, (str, bytes, Promise)):
        return as_list(value)
    if isinstance(value, str):
        if "\n" in value:
            return as_text(value)
        if is_email_simple(value):
            return as_email(value)

    return value


# Formatting functions: never pass None, always return string


def as_email(value):
    return hg.A(value, href=hg.format("mailto: {}", value))


def as_url(value):
    return hg.A(value, href=value, target="_blank", rel="noopener noreferrer")


def as_text(value):
    return mark_safe(linebreaks(value, autoescape=True))


def as_duration(value):
    return str(value - datetime.timedelta(microseconds=value.microseconds))


def as_countries(value):
    return as_list((c.name for c in value))


def as_list(iterable):
    return hg.UL(hg.Iterator(iterable, "item", hg.LI(hg.C("item"))))


def as_download(value, label=None):
    if not value:
        return CONSTANTS[None]
    if not value.storage.exists(value.name):
        return hg.SMALL(hg.EM(_("File not found")))
    if label is None:
        label = hg.SPAN(os.path.basename(value.name))
    return hg.A(
        layout.icon.Icon(
            "launch",
            size=16,
            style="vertical-align: middle; margin-right: 0.25rem;",
        ),
        label,
        newtab=True,
        href=value.url,
        style="margin-right: 0.5rem; margin-left: 0.5rem",
        onclick="event.stopPropagation();",
    )


CONSTANTS = {
    None: getattr(settings, "HTML_NONE", app_settings.HTML_NONE),
    True: getattr(settings, "HTML_TRUE", app_settings.HTML_TRUE),
    False: getattr(settings, "HTML_FALSE", app_settings.HTML_FALSE),
}


# copied from
# https://github.com/django/django/blob/1f9874d4ca3e7376036646aedf6ac3060f22fd69/django/utils/html.py
# because older versions only have this as a local function available
def is_email_simple(value):
    """Return True if value looks like an email address."""
    # An @ must be in the middle of the value.
    if "@" not in value or value.startswith("@") or value.endswith("@"):
        return False
    try:
        p1, p2 = value.split("@")
    except ValueError:
        # value contains more than one @.
        return False
    # Dot must be in p2 (e.g. example.com)
    if "." not in p2 or p2.startswith("."):
        return False
    return True



# Generated by Django 3.2.12 on 2022-04-09 15:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reports', '0009_auto_20220406_1033'),
    ]

    operations = [
        migrations.AlterField(
            model_name='reportcolumn',
            name='sortingname',
            field=models.CharField(blank=True, help_text='Django sorting expression', max_length=255, verbose_name='Sortingname'),
        ),
    ]



import htmlgenerator as hg
from django.contrib.messages.views import SuccessMessageMixin
from django.utils.translation import gettext as _
from django.views.generic import CreateView
from guardian.mixins import PermissionRequiredMixin

from ..utils import pretty_modelname
from .util import BreadView, CustomFormMixin


class AddView(
    CustomFormMixin,
    BreadView,
    SuccessMessageMixin,
    PermissionRequiredMixin,
    CreateView,
):
    """TODO: documentation"""

    accept_global_perms = True

    def get_success_message(self, cleaned_data):
        return _("Added %s") % self.object

    def __init__(self, *args, **kwargs):
        self.fields = kwargs.get("fields", getattr(self, "fields", ["__all__"]))
        super().__init__(*args, **kwargs)

    def get_required_permissions(self, request):
        return [f"{self.model._meta.app_label}.add_{self.model.__name__.lower()}"]

    def get_permission_object(self):
        return None

    def get_context_data(self, *args, **kwargs):
        layout = hg.BaseElement(
            hg.H3(
                _("Add %s") % pretty_modelname(self.model),
            ),
            self._get_layout_cached(),
        )
        return {
            **super().get_context_data(*args, **kwargs),
            "layout": layout,
            "pagetitle": _("Add %s") % pretty_modelname(self.model),
        }



import datetime

import htmlgenerator as hg
from django.core.paginator import Paginator
from django.db import models
from django.http import HttpResponseNotFound
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _

from bread import formatters, layout, views
from bread.utils import filter_fieldlist, generate_excel, xlsxresponse
from bread.utils.links import ModelHref

from .models import Report


class EditView(views.EditView):
    def get_layout(self):
        modelclass = self.object.model.model_class()
        if modelclass is None:
            return layout.notification.InlineNotification(
                "Error",
                f"Model '{self.object.model}' does no longer exist.",
                kind="error",
            )
        column_helper = layout.modal.Modal(
            _("Field explorer"), layout.fieldexplorer.field_help(modelclass), size="lg"
        )

        F = layout.forms.FormField

        fieldstable = layout.forms.FormsetField.as_datatable(
            "columns",
            ["header", "column", "sortingname"],
            formsetfield_kwargs={
                "extra": 1,
                "can_order": True,
            },
        )
        fieldstable[0][1].insert(
            0,
            layout.button.Button(
                _("Help"), buttontype="ghost", **column_helper.openerattributes
            ),
        )
        fieldstable.append(hg.DIV(style="height: 1rem"))
        ret = hg.BaseElement(
            views.header(),
            layout.forms.Form(
                hg.C("form"),
                hg.DIV(
                    _("Base model"),
                    ": ",
                    hg.C("object.model"),
                    style="margin: 2rem 0 2rem 0",
                ),
                F("name"),
                F(
                    "filter",
                    inputelement_attrs={"rows": 1},
                    style="width: 100%",
                ),
                fieldstable,
                layout.tile.ExpandableTile(
                    hg.H4(_("Extended settings")),
                    hg.DIV(
                        F("custom_queryset"),
                        F("pagination"),
                    ),
                ),
                layout.forms.helpers.Submit(style="margin-top: 1rem"),
                column_helper,
            ),
            hg.C("object.preview"),
        )
        return ret

    def get_success_url(self):
        return self.request.get_full_path()


class ReadView(views.ReadView):
    def get_layout(self):

        # ordering, copied from bread.views.browse.BrowseView.get_queryset
        qs = self.object.queryset
        if not qs:
            return layout.notification.InlineNotification(
                "Error",
                f"Model '{self.object.model}' does no longer exist.",
                kind="error",
            )

        order = self.request.GET.get("ordering")
        if order:
            if order.endswith("__int"):
                order = order[: -len("__int")]
                qs = qs.order_by(
                    models.functions.Cast(order[1:], models.IntegerField()).desc()
                    if order.startswith("-")
                    else models.functions.Cast(order, models.IntegerField())
                )
            else:
                qs = qs.order_by(
                    models.functions.Lower(order[1:]).desc()
                    if order.startswith("-")
                    else models.functions.Lower(order)
                )
        paginator = Paginator(qs, self.object.pagination)

        columns = []
        for col in self.object.columns.all():
            sortingname = None
            try:
                sortingname = layout.datatable.sortingname_for_column(
                    self.object.model.model_class(), col.column
                )
            except AttributeError:
                pass
            columns.append(
                layout.datatable.DataTableColumn(
                    col.header,
                    layout.FC(f"row.{col.column}"),
                    col.sortingname or sortingname,
                )
            )
        if not columns:
            columns = [
                layout.datatable.DataTableColumn.from_modelfield(
                    col, self.object.model.model_class()
                )
                for col in filter_fieldlist(
                    self.object.model.model_class(), ["__all__"]
                )
            ]
        pagination_config = layout.pagination.PaginationConfig(
            paginator=paginator,
            items_per_page_options=[self.object.pagination],
        )
        # generate a nice table
        return hg.BaseElement(
            views.header(),
            layout.datatable.DataTable(
                columns=columns,
                row_iterator=paginator.get_page(
                    self.request.GET.get(pagination_config.page_urlparameter)
                )
                if self.object.pagination
                else qs,
            ).with_toolbar(
                title=self.object.name,
                helper_text=f"{self.object.queryset.count()} "
                f"{self.object.model.model_class()._meta.verbose_name_plural}",
                primary_button=layout.button.Button(
                    label=_("Excel"), icon="download"
                ).as_href(ModelHref.from_object(self.object, "excel")),
                pagination_config=pagination_config if self.object.pagination else None,
            ),
        )


def exceldownload(request, pk: int):
    report = get_object_or_404(Report, pk=pk)
    if report.model.model_class() is None:
        return HttpResponseNotFound()

    columns = {
        column.name: lambda row, c=column.column: formatters.format_value(
            hg.resolve_lookup(row, c)
        )
        for column in report.columns.all()
    }
    if not columns:
        columns = {
            column: lambda row, c=column: formatters.format_value(
                hg.resolve_lookup(row, c)
            )
            for column in filter_fieldlist(report.model.model_class(), ["__all__"])
        }

    workbook = generate_excel(report.queryset, columns)
    workbook.title = report.name

    return xlsxresponse(
        workbook, workbook.title + f"-{datetime.date.today().isoformat()}"
    )



import django_countries
from django.db import models
from django.db.models import Q
from django_countries.fields import CountryField


def get_char_text_qset(fields, searchquery, prefix):
    char_text_fields = {
        f
        for f in fields
        if isinstance(f, models.fields.CharField)
        or isinstance(f, models.fields.TextField)
    }

    return {
        Q(**{prefix + "_".join((f.name, "_contains")): searchquery})
        for f in char_text_fields
    }


def get_country_qset(fields, searchquery, prefix):
    countries = {name.lower(): code for code, name in django_countries.countries}
    country_fields = {f for f in fields if isinstance(f, CountryField)}

    if not country_fields:
        return set()

    match_countries = {
        country_name for country_name in countries if searchquery in country_name
    }

    return {
        Q(**{prefix + f.name: countries[key]})
        for f in country_fields
        for key in match_countries
    }


def get_field_queryset(fields, searchquery, prefix="", follow_relationships=1):
    queryset = {
        *get_char_text_qset(fields, searchquery, prefix),
        *get_country_qset(fields, searchquery, prefix),
    }

    qs = Q()
    for query in queryset:
        qs |= query

    if follow_relationships > 0:
        foreignkey_fields = {
            f
            for f in fields
            if isinstance(f, models.fields.related.ForeignKey)
            or isinstance(f, models.fields.related.ManyToManyField)
        }
        for foreignkey_field in foreignkey_fields:
            # skip fields with a name beginning with '_'
            if foreignkey_field.name[0] == "_":
                continue

            if foreignkey_field.related_model:
                foreign_fields = foreignkey_field.related_model._meta.fields
            else:
                foreign_fields = foreignkey_field._meta.fields

            new_prefix = prefix + "__".join([foreignkey_field.name, ""])
            qs |= get_field_queryset(
                foreign_fields, searchquery, new_prefix, follow_relationships - 1
            )

    return qs



# Generated by Django 3.1.7 on 2021-06-19 08:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("reports", "0003_report_custom_queryset"),
    ]

    operations = [
        migrations.AlterField(
            model_name="report",
            name="custom_queryset",
            field=models.CharField(
                blank=True,
                help_text="Key in 'settings.REPORT_FILTERS' must be a function returning a queryset",
                max_length=255,
                verbose_name="Custom Filter",
            ),
        ),
    ]



# Generated by Django 3.2.12 on 2022-03-28 08:34

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('document_templates', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='documenttemplate',
            options={'verbose_name': 'Document Template', 'verbose_name_plural': 'Document Templates'},
        ),
        migrations.AddField(
            model_name='documenttemplate',
            name='model',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.PROTECT, to='contenttypes.contenttype', verbose_name='Model'),
            preserve_default=False,
        ),
        migrations.CreateModel(
            name='DocumentTemplateVariable',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
                ('value', models.CharField(max_length=255, verbose_name='Value')),
                ('raw_value', models.BooleanField(default=False, verbose_name='Raw value')),
                ('document_template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='variables', to='document_templates.documenttemplate')),
            ],
            options={
                'verbose_name': 'Variable',
                'verbose_name_plural': 'Variables',
            },
        ),
    ]



import htmlgenerator as hg
from django.utils.translation import gettext_lazy as _

from .button import Button
from .icon import Icon
from .loading import Loading


class Modal(hg.DIV):
    SIZES = ["xs", "sm", "md", "lg"]

    def __init__(
        self,
        heading,
        *content,
        label="",
        size="sm",
        buttons=(),
        id=None,
        with_form=False,
        **attributes,
    ):
        """
        heading: Modal title
        content: content elements of the modal, can be empty
        label: small informative label above heading
        size: one of ["xs", "sm", "md", "lg"]
        buttons: buttons displayed on bottom of modal, last button has default focus
                 the attribute "data_modal_close" can be set on an button in order to make it a cancel button

        In order to open the modal just pass self.openerattributes as kwargs to another html element, e.g. a button

            modal = modal.Modal("My Modal", "Hello world")
            button.Button("Model", **modal.openerattributes)

        """
        if size not in Modal.SIZES:
            raise ValueError(
                f"argument 'size' has value {size} but needs to be one of {Modal.SIZES}"
            )
        if buttons and "data_modal_primary_focus" not in buttons[-1].attributes:
            buttons[-1].attributes["data_modal_primary_focus"] = True
        attributes["_class"] = attributes.get("_class", "") + " bx--modal"
        self.id = hg.html_id(self, prefix="modal-") if id is None else id
        self.openerattributes = {"data_modal_target": hg.format("#{}", self.id)}
        self.contentcontainer = hg.DIV(
            *content,
            _class="bx--modal-content"
            + (" bx--modal-content--with-form" if with_form else ""),
            tabindex=0,
        )
        super().__init__(
            hg.DIV(
                hg.DIV(
                    hg.P(
                        label,
                        _class="bx--modal-header__label",
                    ),
                    hg.P(
                        heading,
                        _class="bx--modal-header__heading",
                    ),
                    hg.BUTTON(
                        Icon(
                            "close",
                            size=16,
                            _class="bx--modal-close__icon",
                            aria_hidden="true",
                        ),
                        _class="bx--modal-close",
                        type="button",
                        data_modal_close=True,
                    ),
                    _class="bx--modal-header",
                ),
                self.contentcontainer,
                hg.DIV(_class="bx--modal-content--overflow-indicator"),
                hg.DIV(
                    *buttons,
                    _class="bx--modal-footer",
                )
                if buttons
                else "",
                _class=f"bx--modal-container  bx--modal-container--{size}",
            ),
            data_modal=True,
            id=self.id,
            role="dialog",
            aria_modal="true",
            tabindex="-1",
            **attributes,
        )

    @classmethod
    def with_ajax_content(
        cls, heading, url, label="", size="xs", submitlabel=None, id=None, **attributes
    ):
        """
        Same arguments as Modal() except ``url`` replaces ``content`` and ``submitlabel`` replaces ``buttons``

        url: string or htmlgenerator.Lazy
        submitlabel: string or an htmlgenerator element which will be displayed on the submit button.
                     A value of None means no submit button should be displayed.
        """
        buttons = (Button(_("Cancel"), buttontype="ghost", data_modal_close=True),)
        if submitlabel:
            buttons += (Button(submitlabel, type="submit"),)

        modal = cls(
            heading,
            hg.DIV(
                Loading(),
                style="opacity: 0.5; background-color: #EEE; text-align: center;",
            ),
            label="",
            buttons=buttons,
            size=size,
            id=id,
            with_form=bool(submitlabel),
            **attributes,
        )
        if submitlabel:
            buttons[1].attributes["hx_post"] = url
            # note: we always use multipart forms, avoids some issues
            # see ./forms/__init__.py:Form.__init__
            buttons[1].attributes["hx_encoding"] = "multipart/form-data"
            buttons[1].attributes["hx_target"] = hg.format(
                "#{} .bx--modal-content", modal.id
            )
            buttons[1].attributes["hx_include"] = hg.format(
                "#{} .bx--modal-content", modal.id
            )

        modal.openerattributes["hx_get"] = url
        modal.openerattributes["hx_target"] = hg.format(
            "#{} .bx--modal-content", modal.id
        )
        return modal


def modal_with_trigger(modal: Modal, triggerclass: type, *args, **kwargs):
    return hg.BaseElement(
        triggerclass(*args, **{**kwargs, **modal.openerattributes}), modal
    )



from celery import Task


class RepeatedTask(Task):
    @classmethod
    def on_bound(cls, app):
        app.conf.beat_schedule[cls.name] = {
            "task": cls.name,
            "schedule": cls.run_every.total_seconds(),
            "args": (),
            "kwargs": {},
            "options": getattr(cls, "options", {}),
            "relative": getattr(cls, "relative", False),
        }



from django.utils.translation import gettext_lazy as _
from dynamic_preferences.preferences import Section
from dynamic_preferences.types import StringPreference
from dynamic_preferences.users.registries import user_preferences_registry

general = Section("general", _("General"))


@user_preferences_registry.register
class PreferredLanguage(StringPreference):
    section = general
    name = "preferred_language"
    default = ""
    verbose_name = _("Preferred Language")


@user_preferences_registry.register
class Timezone(StringPreference):
    section = general
    name = "timezone"
    default = ""
    verbose_name = _("Timezone")



from typing import List, NamedTuple, Optional, Union

import htmlgenerator as hg
from django.db import models

from .urls import model_urlname
from .urls import reverse as urlreverse


class LazyHref(hg.Lazy):
    """An element which will resolve lazy. The ``args`` and ``kwargs`` arguments will
    be passed to ``bread.utils.urls.reverse``. Every (lazy) item in ``args`` will be
    resolved and every value in ``kwargs`` will be resolved.

    Example usage:

        assert "/settings" == LazyHref(hg.F(lambda c: "settings"))

    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def resolve(self, context: dict):
        kwargs = {k: hg.resolve_lazy(v, context) for k, v in self.kwargs.items()}
        # the django reverse function requires url-keyword arguments to be pass
        # in a parameter named "kwarg". This is a bit confusing since kwargs
        # normally referse to the python keyword arguments and not to URL
        # keyword arguments. However, we also want to support lazy URL
        # keywords, so we do the resolving of the actualy URL-kwargs as well
        if "kwargs" in kwargs:
            kwargs["kwargs"] = {
                k: hg.resolve_lazy(v, context) for k, v in kwargs["kwargs"].items()
            }
        if "args" in kwargs:
            kwargs["args"] = [hg.resolve_lazy(arg, context) for arg in kwargs["args"]]
        if "query" in kwargs:
            kwargs["query"] = {
                k: hg.resolve_lazy(v, context) for k, v in kwargs["query"].items()
            }
        return urlreverse(*[hg.resolve_lazy(a, context) for a in self.args], **kwargs)


class ModelHref(LazyHref):
    """
    Works similar to LazyHref but takes a model and a model action
    like "edit", "read", "browse" and generates the URL automatically
    (according to bread conventions).
    This is usefull as a replacment of wrapping ``bread.utils.urls.reverse_model``
    inside a hg.F element for lazy evalutation of the pk value.

    Example usage:

        assert "/person/browse" == ModelHref(models.Person, "browse").resolve(context)
        assert "/person/edit/1" == ModelHref(
            models.Person, "edit", kwargs={"pk": hg.C("object.pk")}
        ).resolve(context)

    return_to_current: will add a URL query parameter "next=<current_url" to the generated URL

    """

    def __init__(
        self,
        model: Union[models.Model, hg.Lazy],
        name: str,
        *args,
        return_to_current: bool = False,
        **kwargs
    ):

        # if this is an instance of a model, we can extract the pk URL argument directly
        # TODO: instance-specific routes which don't use the pk argument will fail
        if isinstance(model, hg.Lazy):
            url = hg.F(lambda c: model_urlname(hg.resolve_lazy(model, c), name))
        else:
            url = model_urlname(model, name)
        if return_to_current:
            if "query" not in kwargs:
                kwargs["query"] = {}
            kwargs["query"]["next"] = hg.C("request.get_full_path")

        super().__init__(url, *args, **kwargs)

    @staticmethod
    def from_object(object: Union[models.Model, hg.Lazy], name: str, *args, **kwargs):
        """
        name: string which denotes an object action name as generated
              by bread.utils.urls.model_urlname
        object: instance of a Model, can be lazy
        """

        if "kwargs" not in kwargs:
            kwargs["kwargs"] = {}
        kwargs["kwargs"]["pk"] = object.pk
        return ModelHref(object, name, *args, **kwargs)


def try_call(var, *args, **kwargs):
    return var(*args, **kwargs) if callable(var) else var


class Link(NamedTuple):
    href: Union[str, hg.Lazy]
    label: str
    iconname: Optional[str] = "fade"
    permissions: List[str] = []
    attributes: dict = {}

    def has_permission(self, request, obj=None):
        return all(
            [
                request.user.has_perm(perm, obj) or request.user.has_perm(perm)
                for perm in self.permissions
            ]
        )



import htmlgenerator as hg
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from bread import layout


def generate_wizard_form(wizardview, wizardtitle, steptitle, formlayout):
    """
    title: Title of the current page
    steps: list of 2-tuples with (step_title, status) where status must be one of ["incomplete", "complete", "current"]
    """
    # needs to be rendered in view of type NamedUrlSessionWizardView in order to work correctly
    def go_back_url(context):
        url = reverse(
            context["request"].resolver_match.view_name,
            kwargs={"step": context["wizard"]["steps"].prev},
        )
        return f"document.location='{url}'"

    steps = []
    for i, (step, formclass) in enumerate(wizardview.get_form_list().items()):
        status = "incomplete"
        if i < wizardview.steps.index:
            status = "complete"
        if step == wizardview.steps.current:
            status = "current"
        steps.append((formclass.title, status))

    return hg.BaseElement(
        hg.H3(wizardtitle),
        hg.H4(steptitle),
        layout.progress_indicator.ProgressIndicator(
            steps,
            style="margin-bottom: 2rem",
        ),
        layout.forms.Form(
            hg.C("wizard.form"),
            layout.forms.Form(
                hg.C("wizard.management_form"),
                layout.forms.FormField("current_step", form="wizard.management_form"),
                standalone=False,
            ),
            formlayout,
            hg.DIV(
                hg.DIV(
                    hg.If(
                        hg.C("wizard.steps.prev"),
                        layout.button.Button(
                            _("Back"),
                            buttontype="secondary",
                            onclick=hg.F(go_back_url),
                        ),
                    ),
                    hg.If(
                        hg.F(
                            lambda c: c["wizard"]["steps"].last
                            == c["wizard"]["steps"].current
                        ),
                        layout.button.Button(
                            _("Complete"), type="submit", style="margin-left: 1rem"
                        ),
                        layout.button.Button(
                            _("Continue"), type="submit", style="margin-left: 1rem"
                        ),
                    ),
                ),
                style="align-items: flex-end",
                _class="bx--form-item",
            ),
        ),
    )



# Generated by Django 3.2.12 on 2022-04-06 10:33

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('reports', '0007_report_pagination'),
    ]

    operations = [
        migrations.RenameField(
            model_name='reportcolumn',
            old_name='name',
            new_name='header',
        ),
    ]



from django import forms
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.utils.html import strip_tags


class GenericForeignKeyField(forms.TypedChoiceField):
    @classmethod
    def object_to_choice(cls, obj):
        return (
            f"{ContentType.objects.get_for_model(obj).pk},{obj.pk}",
            f"{obj._meta.verbose_name.title()}: {strip_tags(str(obj))}",
        )

    @classmethod
    def objects_to_choices(cls, objects, required=True):
        yield None, "---"
        for obj in objects:
            yield GenericForeignKeyField.object_to_choice(obj)

    def __init__(self, **kwargs):
        def convert(value):
            contentype_id, object_id = map(int, value.split(","))
            return ContentType.objects.get(pk=contentype_id).get_object_for_this_type(
                pk=object_id
            )

        super().__init__(coerce=convert, empty_value=None, **kwargs)


class BoundFormsetField(forms.BoundField):
    @property
    def formset(self):
        prefix = self.field.formsetclass.get_default_prefix()
        if self.form.prefix:
            prefix = f"{self.form.prefix}-{prefix}"
        return self.field.formsetclass(
            self.form.data if self.form.is_bound else None,
            self.form.files if self.form.is_bound else None,
            instance=self.field.parent_instance,
            prefix=prefix,
            **(self.field.formsetargs or {}),
        )

    @property
    def data(self):
        return self.formset


class FormsetField(forms.Field):
    def __init__(
        self, formsetclass, parent_instance, formsetargs=None, *args, **kwargs
    ):
        self.widget = FormsetWidget(formsetclass, parent_instance, formsetargs)
        self.formsetclass = formsetclass
        self.parent_instance = parent_instance
        self.formsetargs = formsetargs
        kwargs["required"] = False
        kwargs.setdefault("label", formsetclass.model._meta.verbose_name_plural)
        super().__init__(*args, **kwargs)

    def to_python(self, value):
        return self.formsetclass(**value)

    def validate(self, value):
        super().validate(value)
        if not value.is_valid():
            for formerrors in value.errors:
                for field, errorlist in formerrors.items():
                    for error in errorlist:
                        if isinstance(error, str):
                            error = ValidationError(error)
                        raise error

    def get_bound_field(self, form, field_name):
        return BoundFormsetField(form, self, field_name)


class FormsetWidget(forms.Widget):
    def __init__(
        self, formsetclass, parent_instance, *args, formsetargs=None, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.formsetclass = formsetclass
        self.formsetargs = formsetargs
        self.parent_instance = parent_instance
        self.needs_multipart_form = self.formsetclass().is_multipart()
        self.template_name = "django/forms/widgets/text.html"  # we set this in order to prevent __str__ on the parent form to fail

    def value_from_datadict(self, data, files, name):
        return {
            "data": data,
            "files": files,
            "instance": self.parent_instance,
            "prefix": name,
            **(self.formsetargs or {}),
        }



# Generated by Django 3.2.12 on 2022-03-28 09:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('document_templates', '0002_auto_20220328_0834'),
    ]

    operations = [
        migrations.AlterField(
            model_name='documenttemplate',
            name='file',
            field=models.FileField(help_text="Must be a *.docx file, use '{{ variable-name }}' to insert variables from below", upload_to='document_templates/'),
        ),
        migrations.AlterField(
            model_name='documenttemplatevariable',
            name='name',
            field=models.CharField(help_text='Name to use in the template document', max_length=255, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='documenttemplatevariable',
            name='value',
            field=models.CharField(help_text='Path to the desired value (see help)', max_length=255, verbose_name='Value'),
        ),
    ]



import datetime

import htmlgenerator as hg
from django.conf import settings
from django.db import models
from django.db.models import ManyToOneRel
from django.template.defaultfilters import linebreaksbr
from django.utils.formats import localize
from django.utils.timezone import localtime

from ..formatters import format_value
from ..utils import pretty_modelname, resolve_modellookup, reverse_model

DEVMODE_KEY = "DEVMODE"


class HasBreadCookieValue(hg.Lazy):
    def __init__(self, cookiename, value, default=None):
        self.cookiename = cookiename
        self.value = value
        self.default = default

    def resolve(self, context):
        if f"bread-{self.cookiename}" in context["request"].session.get(
            "bread-cookies", {}
        ):
            return (
                context["request"].session["bread-cookies"][f"bread-{self.cookiename}"]
                == self.value
            )
        return self.default == self.value


class DevModeOnly(hg.BaseElement):
    def render(self, context):
        if context["request"].session.get(DEVMODE_KEY, False):
            return super().render(context)
        return ""


def objectaction(object, action, *args, **kwargs):
    kwargs["kwargs"] = {"pk": object.pk}
    return str(
        reverse_model(
            object,
            action,
            *args,
            **kwargs,
        )
    )


def aslink_attributes(href):
    """
    Shortcut to generate HTMLElement attributes to make any element behave like a link.
    This should normally be used like this: hg.DIV("hello", \\*\\*aslink_attributes('google.com'))
    """
    return {
        "onclick": hg.BaseElement("document.location = '", href, "'"),
        "onauxclick": hg.BaseElement("window.open('", href, "', '_blank')"),
        "style": "cursor: pointer",
    }


class ModelName(hg.ContextValue):
    def resolve(self, context):
        return pretty_modelname(super().resolve(context))


class FormattedContextValue(hg.ContextValue):
    def resolve(self, context):
        return str(format_value(super().resolve(context)))


class ObjectFieldLabel(hg.ContextValue):
    def __init__(self, fieldname, object_contextname="object", title=True):
        """
        :param fieldname: Name of the model field whose value will be rendered
        :param object_contextname: Name of the context object which provides the field value
                                   or a Lazy which produces the object itself
        """
        super().__init__(object_contextname)
        self.fieldname = fieldname
        self.title = title
        self.object = object_contextname

    def resolve(self, context):
        object = self.object
        if isinstance(self.object, str):
            object = resolve_modellookup(context, self.object)[0]
        object = hg.resolve_lazy(object, context)
        label = resolve_modellookup(object._meta.model, self.fieldname)[-1]
        if hasattr(label, "verbose_name"):
            return label.verbose_name
        if isinstance(label, property):
            return label.fget.__name__.replace("_", " ")
        if callable(label):
            return label.__name__.replace("_", " ")
        if isinstance(label, ManyToOneRel):
            return (
                label.related_model._meta.verbose_name_plural
            )  # this is "more correct", but not sure if it always works..
            # return label.name.replace("_", " ").capitalize()
        return label.title() if self.title and isinstance(label, str) else label


class ObjectFieldValue(hg.Lazy):
    def __init__(self, fieldname, object_contextname="object", formatter=None):
        """
        :param fieldname: Name of the model field whose value will be rendered
        :param object_contextname: Name of the context object which provides the field value
                                   or a Lazy which produces the object itself
        :param formatter: function which takes the field value as a single
                          argument and returns a formatted version
        """
        self.object = object_contextname
        self.fieldname = fieldname
        self.formatter = formatter

    def resolve(self, context):
        object = self.object
        if isinstance(self.object, str):
            object = resolve_modellookup(context, self.object)[0]
        object = hg.resolve_lazy(object, context)

        parts = self.fieldname.split(".")
        # test if the value has a matching get_FIELDNAME_display function
        try:
            value = hg.resolve_lookup(
                object, f"{'.'.join(parts[:-1])}.get_{parts[-1]}_display".lstrip(".")
            )
        except Exception:
            value = None
        if value is None:
            try:
                value = hg.resolve_lookup(object, self.fieldname)
            except AttributeError:
                # e.g. for non-existing OneToOneField related value
                pass
        if isinstance(value, datetime.datetime):
            value = localtime(value)
        if self.formatter:
            value = self.formatter(value)
        value = localize(value, use_l10n=settings.USE_L10N)
        if isinstance(value, models.Manager):
            value = ", ".join([str(x) for x in value.all()])
        if isinstance(value, str):
            value = linebreaksbr(value)
        return value


FC = FormattedContextValue



from collections import namedtuple

from django.contrib.auth.models import User
from django.db.models.fields.proxy import OrderWrt
from django.test import Client
from hypothesis import given, settings
from hypothesis.extra.django import TestCase, from_model, register_field_strategy
from hypothesis.strategies import integers

from bread.utils.urls import reverse_model

FormField = namedtuple("FormField", ["name", "value", "element", "generators"])


register_field_strategy(OrderWrt, integers(-2147483648, 2147483647))


def authenticate(client):
    user, _ = User.objects.get_or_create(username="admin", is_superuser=True)
    client.force_login(user)


def generic_bread_testcase(model, **kwargs):
    class GenericModelTest(TestCase):
        def setUp(self):
            self.client = Client()
            authenticate(self.client)

        @given(from_model(model, **kwargs))
        def test_generic(self, instance):
            self.assertIsNotNone(instance.pk)
            resp = self.client.get(reverse_model(model, "browse"), follow=True)
            self.assertEqual(resp.status_code, 200)
            resp = self.client.get(
                reverse_model(model, "read", kwargs={"pk": instance.pk}), follow=True
            )
            self.assertEqual(resp.status_code, 200)
            resp = self.client.get(
                reverse_model(model, "edit", kwargs={"pk": instance.pk}), follow=True
            )
            resp = self.client.get(
                reverse_model(model, "delete", kwargs={"pk": instance.pk}), follow=True
            )
            self.assertEqual(resp.status_code, 200)

        def test_add(self):
            resp = self.client.get(reverse_model(model, "add"), follow=True)
            self.assertEqual(resp.status_code, 200)
            # TODO: implement add form, use hypothesis

        @given(from_model(model, **kwargs))
        @settings(deadline=None)
        def test_forms(self, instance):
            self.assertIsNotNone(instance.pk)
            # TODO: implement edit form, use hypothesis

    return GenericModelTest



import htmlgenerator as hg
from django.utils.translation import gettext_lazy as _

from bread import layout, utils, views


class WorkflowBrowseView(views.BrowseView):
    rowclickaction = views.BrowseView.gen_rowclickaction("edit")

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .exclude(completed__isnull=False)
            .exclude(cancelled__isnull=False)
        )

    def get_layout(self):
        workflow_diagram = layout.modal.Modal(
            _("Workflow Diagram"),
            self.model.workflow_as_svg(),
            label=self.model._meta.verbose_name,
        )

        ret = super().get_layout()
        ret[0][0].append(
            layout.button.Button(
                _("Explain Workflow"),
                **workflow_diagram.openerattributes,
                small=True,
                style="float: right",
                buttontype="tertiary",
            ),
        )
        ret.append(workflow_diagram)
        return ret


class WorkflowEditView(views.EditView):
    def get_layout(self):
        fields = hg.BaseElement(
            *[layout.forms.FormField(f) for f in self.object.active_fields()]
        )
        return hg.BaseElement(
            hg.H1(self.object, style="margin-bottom: 2rem"),
            hg.DIV(
                hg.DIV(
                    layout.forms.Form(
                        hg.C("form"), fields, layout.forms.helpers.Submit()
                    ),
                    style="padding: 1rem",
                ),
                hg.DIV(
                    self.object.as_svg(), style="width: 40%; border: 1px solid gray"
                ),
                style="display: flex",
            ),
        )


class WorkflowReadView(views.ReadView):
    def get_layout(self):
        fields = hg.BaseElement(
            *[
                layout.forms.FormField(f)
                for f in utils.filter_fieldlist(self.model, ["__all__"], for_form=True)
            ]
        )

        return hg.BaseElement(
            hg.H1(self.object, style="margin-bottom: 2rem"),
            hg.DIV(
                layout.button.Button(
                    "Edit",
                    **layout.aslink_attributes(
                        layout.objectaction(self.object, "edit")
                    ),
                ),
                views.layoutasreadonly(
                    hg.DIV(
                        hg.DIV(
                            layout.forms.Form(
                                hg.C("form"), fields, layout.forms.helpers.Submit()
                            ),
                            style="padding: 1rem",
                        ),
                        hg.DIV(
                            self.object.as_svg(),
                            style="width: 40%; border: 1px solid gray",
                        ),
                        style="display: flex;",
                    )
                ),
            ),
        )


class WorkflowEditFullView(views.EditView):
    def get_layout(self):
        return hg.DIV(
            hg.DIV(
                super().get_layout(),
                style="flex-grow: 1",
            ),
            hg.DIV(self.object.as_svg(), style="width: 40%"),
            style="display: flex",
        )



from django.apps import AppConfig
from django.contrib.messages.constants import DEFAULT_TAGS
from django.utils.translation import gettext as _


class BreadConfig(AppConfig):

    name = "bread"
    verbose_name = "Bread Engine"

    def ready(self):
        # trigger translation of message tags
        [_(tag.capitalize()) for tag in DEFAULT_TAGS.values()]



from django.conf import settings
from django.http import HttpResponse
from django.template.context import _builtin_context_processors
from django.utils.module_loading import import_string

from . import components  # noqa
from .componentpreview import *  # noqa
from .components import *  # noqa
from .components import button  # noqa
from .components import content_switcher  # noqa
from .components import datatable  # noqa
from .components import fieldexplorer  # noqa
from .components import forms  # noqa
from .components import grid  # noqa
from .components import icon  # noqa
from .components import loading  # noqa
from .components import modal  # noqa
from .components import notification  # noqa
from .components import overflow_menu  # noqa
from .components import pagination  # noqa
from .components import progress_indicator  # noqa
from .components import search  # noqa
from .components import shell_header  # noqa
from .components import sidenav  # noqa
from .components import tabs  # noqa
from .components import tag  # noqa
from .components import toggle  # noqa
from .components import tooltip  # noqa
from .components.forms import helpers  # noqa
from .components.forms import search_select  # noqa
from .skeleton import *  # noqa
from .utils import *  # noqa


def render(request, layout, context=None, **response_kwargs):
    if render.CONTEXT_PROCESSORS is None:
        render.CONTEXT_PROCESSORS = tuple(
            import_string(path)
            for path in _builtin_context_processors
            + tuple(
                (settings.TEMPLATES + [{}])[0]
                .get("OPTIONS", {})
                .get("context_processors", [])
            )
        )
    response_kwargs.setdefault("content_type", "text/html")
    defaultcontext = {}
    for processor in render.CONTEXT_PROCESSORS:
        defaultcontext.update(processor(request))
    defaultcontext.update(context or {})
    return HttpResponse(layout.render(defaultcontext), **response_kwargs)


render.CONTEXT_PROCESSORS = None  # type: ignore



import html
import io
import re

import htmlgenerator as hg
from django.conf import settings
from django.http import HttpResponse
from django.utils.html import strip_tags


def html_to_pdf(html, as_http_response=False, name=None):
    """
    Renders html to PDF

        html: html string
        as_http_response: If True return PDF as HttpResponse with attachment
        name: Filename without extension, only used when as_http_response == True
        returns: PDF content or HttpResponse with attachment
    """
    # weasyprint is an extra dependency
    import weasyprint

    def url_fetcher(url, timeout=10, ssl_context=None):
        return weasyprint.default_url_fetcher(
            "file://" + url.replace("file://", settings.BASE_DIR), timeout, ssl_context
        )

    ret = weasyprint.HTML(string=html, base_url="", url_fetcher=url_fetcher).write_pdf()
    if as_http_response:
        ret = HttpResponse(ret, content_type="application/pdf")
        ret["Content-Disposition"] = f"inline; filename={name}.pdf"
    return ret


def generate_excel(rows, columns):
    """
    columns: dict with {<columnname>: formatting_function(row)}
    """
    import openpyxl
    from openpyxl.styles import Font

    workbook = openpyxl.Workbook()
    workbookcolumns = workbook.active.iter_cols(
        min_row=1, max_col=len(columns) + 1, max_row=len(rows) + 1
    )
    newline_regex = re.compile(
        r"<\s*br\s*/?\s*>"
    )  # replace HTML line breaks with newlines
    for columnname, columndata in zip(columns, workbookcolumns):
        columndata[0].value = html.unescape(
            strip_tags(newline_regex.sub(r"\n", str(columnname)))
        )
        columndata[0].font = Font(bold=True)
        for i, cell in enumerate(columndata[1:]):
            value = columns[columnname](rows[i])
            if isinstance(value, hg.BaseElement):
                value = hg.render(value, {})
            cleaned = html.unescape(
                strip_tags(newline_regex.sub(r"\n", str(value or "")))
            )
            cell.value = cleaned
    return workbook


def xlsxresponse(workbook, title, filters=True):
    """
    Returns workbook as a downloadable file

        workbook: openpyxl workbook
        title: filename without extension
        returns: HttpResponse with attachment
    """
    workbook = prepare_excel(workbook, filters)
    buf = io.BytesIO()
    workbook.save(buf)
    buf.seek(0)
    response = HttpResponse(buf.read(), content_type="application/vnd.ms-excel")
    response["Content-Disposition"] = f'attachment; filename="{title}.xlsx"'
    return response


def prepare_excel(workbook, filters=True):
    """
    Formats the excel a bit in order to be displayed nicely

        workbook: openpyxl workbook
        filters: If True enable excel filtering headers
        returns: formated workbook
    """
    # openpyxl is an extra requirement
    from openpyxl.styles import Alignment

    for worksheet in workbook:
        # estimate column width
        for col in worksheet.columns:
            max_length = 0
            column = col[0].column_letter
            for cell in col:
                if len(str(cell.value)) > max_length:
                    max_length = len(str(cell.value))
            adjusted_width = (max_length + 3) * 1.2
            worksheet.column_dimensions[column].width = min([adjusted_width, 50])

            # enable excel filters
            if filters is True:
                worksheet.auto_filter.ref = f"A1:{column}1"

        # enable word wrap
        for row in worksheet.iter_rows():
            for cell in row:
                cell.alignment = Alignment(wrap_text=True)
                if isinstance(cell.value, str):
                    cell.value = cell.value.strip()
                    if cell.value.isdigit():
                        cell.value = int(cell.value)

    return workbook



# Generated by Django 3.2.13 on 2022-05-19 10:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('triggers', '0002_auto_20220208_1215'),
    ]

    operations = [
        migrations.AddField(
            model_name='datachangetrigger',
            name='description',
            field=models.CharField(default='Missing description', max_length=255),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='datefieldtrigger',
            name='description',
            field=models.CharField(default='Missing description', max_length=255),
            preserve_default=False,
        ),
    ]



import django_celery_results.models
from django.apps import apps
from django.urls import include, path

from bread.utils import autopath, default_model_paths

from .views import administration, auth, userprofile
from .views.globalpreferences import PreferencesView

external_urlpatterns = [
    path(
        "preferences/",
        include(
            (
                [
                    path("global/", PreferencesView.as_view(), name="global"),
                    path(
                        "global/<slug:section>",
                        PreferencesView.as_view(),
                        name="global.section",
                    ),
                ],
                "dynamic_preferences",
            ),
            namespace="preferences",
        ),
    )
]
urlpatterns = [
    path(
        "accounts/login/",
        auth.BreadLoginView.as_view(),
        name="login",
    ),
    path(
        "accounts/logout/",
        auth.BreadLogoutView.as_view(),
        name="logout",
    ),
    path(
        "accounts/user/",
        userprofile.UserProfileView.as_view(),
        name="userprofile",
    ),
    autopath(
        userprofile.EditPersonalDataView.as_view(), urlname="userprofile.personal"
    ),
    autopath(userprofile.EditLoginView.as_view(), urlname="userprofile.login"),
    autopath(userprofile.password_reset, urlname="userprofile.password_reset"),
    autopath(
        userprofile.EditPermissionsView.as_view(), urlname="userprofile.permissions"
    ),
    autopath(userprofile.set_devmode, urlname="devmode"),
    path(
        "accounts/password_reset/",
        auth.BreadPasswordResetView.as_view(),
        name="password_reset",
    ),
    path(
        "accounts/password_reset/done/",
        auth.BreadPasswordResetDoneView.as_view(),
        name="password_reset_done",
    ),
    path(
        "accounts/reset/<uidb64>/<token>/",
        auth.BreadPasswordResetConfirmView.as_view(),
        name="password_reset_confirm",
    ),
    path(
        "accounts/reset/done/",
        auth.BreadPasswordResetCompleteView.as_view(),
        name="password_reset_complete",
    ),
    path(
        "systeminformation", administration.systeminformation, name="systeminformation"
    ),
    path(
        "admin/maintenance",
        administration.maintenancesettings,
        name="breadadmin.maintenance",
    ),
    path(
        "admin/componentpreview",
        administration.componentpreview,
        name="breadadmin.componentpreview",
    ),
    *default_model_paths(
        django_celery_results.models.TaskResult,
        browseview=administration.TaskResultBrowseView,
    ),
] + external_urlpatterns

for app in apps.get_app_configs():
    if app.name.startswith("bread.contrib"):
        urlpatterns.append(path("contrib/", include(f"{app.name}.urls")))



# Generated by Django 3.2.12 on 2022-04-06 10:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reports', '0006_alter_report_model'),
    ]

    operations = [
        migrations.AddField(
            model_name='report',
            name='pagination',
            field=models.PositiveIntegerField(default=0, help_text='How many items to display per page when viewing the report in the browser, 0 for everything on one page', verbose_name='Pagination'),
        ),
    ]



import datetime
import io
from typing import Union

import htmlgenerator as hg
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import formats
from django.utils.dateformat import DateFormat
from django.utils.timezone import now
from django.utils.translation import get_language
from django.utils.translation import gettext_lazy as _
from docxtpl import DocxTemplate
from jinja2.sandbox import SandboxedEnvironment

from bread.utils import ModelHref


class DocumentTemplate(models.Model):
    name = models.CharField(_("Name"), max_length=255)
    file = models.FileField(
        upload_to="document_templates/",
        help_text=_(
            "Must be a *.docx file, use '{{ variable-name }}' to insert variables from below"
        ),
    )

    model = models.ForeignKey(
        ContentType, on_delete=models.PROTECT, verbose_name=_("Model")
    )

    def default_context(self):
        return {"now": DateFormat(now())}

    def render_with(self, object):
        def ensure_localized_date(value):
            if isinstance(value, (datetime.datetime, datetime.date)):
                return DateFormat(value)
            return value

        docxtpl_template = DocxTemplate(self.file.path)
        context = {}
        for variable in self.variables.all():
            context[variable.name] = ensure_localized_date(
                hg.resolve_lookup(object, variable.value)
            )
            if variable.template:
                value_env = SandboxedEnvironment()
                value_env.filters["map"] = lambda value, map: map.get(value, value)
                try:
                    context[variable.name] = value_env.from_string(
                        variable.template
                    ).render(
                        value=context[variable.name],
                    )
                except Exception as e:
                    context[variable.name] = f"### ERROR: {e} ###"
            else:
                if isinstance(context[variable.name], DateFormat):
                    context[variable.name] = context[variable.name].format(
                        formats.get_format("DATE_FORMAT", lang=get_language())
                    )

        context.update(self.default_context())
        env = SandboxedEnvironment()
        env.filters["map"] = lambda value, map: map.get(value, value)
        docxtpl_template.render(context, env)

        buf = io.BytesIO()
        docxtpl_template.save(buf)
        buf.seek(0)
        return buf

    def missing_variables(self):
        """Returns (variables_only_in_template, variables_only_in_definition)"""
        intemplate = DocxTemplate(
            self.file.path
        ).get_undeclared_template_variables() - set(self.default_context().keys())
        declared = {v.name for v in self.variables.all()}
        both = intemplate | declared
        return declared ^ both, intemplate ^ both

    def generate_document_url(self, obj: Union[hg.Lazy, models.Model]):
        return ModelHref.from_object(
            self,
            "generate_document",
            kwargs={"object_pk": obj.pk},
        )

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _("Document Template")
        verbose_name_plural = _("Document Templates")


class DocumentTemplateVariable(models.Model):
    document_template = models.ForeignKey(
        DocumentTemplate, on_delete=models.CASCADE, related_name="variables"
    )
    name = models.CharField(
        _("Name"), max_length=255, help_text=_("Name to use in the template document")
    )
    value = models.CharField(
        _("Value"), max_length=255, help_text=_("Path to the desired value (see help)")
    )
    template = models.TextField(
        _("Template"),
        blank=True,
        help_text=_("Jinja template with 'value' in context"),
    )
    raw_value = models.BooleanField(_("Raw value"), default=False)

    class Meta:
        verbose_name = _("Variable")
        verbose_name_plural = _("Variables")



import sys
from typing import Union

import htmlgenerator as hg
from django.utils.translation import gettext_lazy as _

from bread.layout.components.button import Button
from bread.utils import Link, aslayout


@aslayout
def error_layout(
    request,
    status_code: int,
    status_title: str,
    description: Union[str, hg.BaseElement],
    exception_detail: str = None,
):
    return hg.BaseElement(
        hg.H1(f"{status_code}: {status_title}", style="margin-bottom: 1rem;"),
        hg.P(
            description,
            style="margin-bottom: 1rem;",
        ),
        hg.If(
            bool(exception_detail),
            hg.BaseElement(
                hg.H4("Detail", style="margin-bottom: 1rem;"),
                hg.DIV(
                    exception_detail,
                    style=(
                        "border: 1px solid grey;"
                        "padding: 1rem;"
                        "font-family: monospace;"
                        "margin-bottom: 1rem;"
                    ),
                ),
            ),
        ),
        Button.from_link(
            Link(
                label=_("Back to homepage"),
                href=hg.F(lambda c: c["request"].META["SCRIPT_NAME"] or "/"),
            )
        ),
    )


def view400(request, exception):
    response = error_layout(
        request,
        status_code=400,
        status_title=_("Bad request"),
        description=_("The server cannot process the request due to a client error."),
        exception_detail=exception.message if hasattr(exception, "message") else None,
    )
    response.status_code = 400

    return response


def view403(request, exception):
    response = error_layout(
        request,
        status_code=403,
        status_title=_("Forbidden"),
        description=_("You do not have permission to access this resource."),
        exception_detail=exception.message if hasattr(exception, "message") else None,
    )
    response.status_code = 403

    return response


def view404(request, exception):
    response = error_layout(
        request,
        status_code=404,
        status_title=_("Page not found"),
        description=hg.BaseElement(
            hg.mark_safe(
                _("The path %s could not be found.")
                % f"<strong>{request.path}</strong>"
            )
        ),
        exception_detail=exception.message if hasattr(exception, "message") else None,
    )
    response.status_code = 404

    return response


def view500(request):
    exec_info = sys.exc_info()
    response = error_layout(
        request,
        status_code=500,
        status_title=_("Internal Server Error"),
        description=hg.BaseElement(
            hg.mark_safe(
                _(
                    "A problem has occurred while loading the page at %s."
                    "<br>If the problem persists, please contact an administrator."
                )
                % f"<strong>{request.path}</strong>"
            )
        ),
        exception_detail=f"{str(exec_info[0].__name__)}: {exec_info[1]}",
    )
    response.status_code = 500

    return response



import urllib

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo  # type: ignore

from django.contrib.auth import authenticate, login
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import activate, get_language


class RequireAuthenticationMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if "bread-cookies" not in request.session:
            request.session["bread-cookies"] = {}
        for k, v in request.COOKIES.items():
            if k.startswith("bread-"):
                request.session["bread-cookies"][k] = urllib.parse.unquote(v)

        # load the preferred language for a user and change translation system
        # and language cookie if required
        if request.user and hasattr(request.user, "preferences"):
            lang = request.user.preferences.get("general__preferred_language")
            if lang and lang != get_language():
                # needs to be called before every request to do translation correctly
                activate(lang)
            tz = request.user.preferences.get("general__timezone")
            if tz:
                timezone.activate(zoneinfo.ZoneInfo(tz))
            else:
                timezone.deactivate()

        return self.get_response(request)

    def process_view(self, request, view_func, view_args, view_kwargs):
        whitelisted_urlnames = (
            "login",
            "password_reset",
            "password_reset_done",
            "password_reset_confirm",
            "password_reset_complete",
        )
        if request.user.is_authenticated:
            return None
        if request.resolver_match.url_name in whitelisted_urlnames:
            return None
        if "Authorization" in request.headers:
            user = authenticate(request)
            if user is not None:
                login(request, user)
                request.user = user
                return None

        return HttpResponseRedirect(reverse("login") + "?next=" + request.path)



# Generated by Django 3.1.5 on 2021-03-01 14:46

import django.db.models.deletion
from django.db import migrations, models

import bread.querysetfield


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Report",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateField(auto_now_add=True, verbose_name="Created"),
                ),
                ("name", models.CharField(max_length=255, verbose_name="Name")),
                (
                    "filter",
                    bread.querysetfield.QuerysetField(
                        blank=True, modelfieldname="model", verbose_name="Filter"
                    ),
                ),
                (
                    "model",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Report",
                "verbose_name_plural": "Reports",
                "ordering": ["name"],
            },
        ),
        migrations.CreateModel(
            name="ReportColumn",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("column", models.CharField(max_length=255, verbose_name="Column")),
                ("name", models.CharField(max_length=255, verbose_name="Name")),
                (
                    "aggregation",
                    models.CharField(
                        blank=True,
                        choices=[("count", ""), ("sum", "")],
                        max_length=64,
                        verbose_name="Aggregation",
                    ),
                ),
                (
                    "report",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="columns",
                        to="reports.report",
                    ),
                ),
            ],
        ),
    ]



import re
import urllib
import urllib.parse
from typing import Iterable, List, Optional, Tuple, Type, Union

import htmlgenerator as hg
from django.contrib import messages
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.translation import gettext as _
from django.utils.translation import pgettext_lazy
from django.views.decorators.cache import never_cache
from django.views.generic import UpdateView
from guardian.mixins import PermissionRequiredMixin

from ..utils import filter_fieldlist, model_urlname, reverse_model
from .util import BreadView, CustomFormMixin


class EditView(
    CustomFormMixin,
    BreadView,
    messages.views.SuccessMessageMixin,
    PermissionRequiredMixin,
    UpdateView,
):
    """TODO: documentation"""

    accept_global_perms = True
    fields: Optional[List[Union[hg.BaseElement, str]]] = None
    urlparams: Iterable[Tuple[str, Type]] = (("pk", int),)

    def __init__(self, *args, **kwargs):
        all = filter_fieldlist(
            kwargs.get("model", getattr(self, "model")), ["__all__"], for_form=True
        )
        self.fields = kwargs.get("fields", getattr(self, "fields", None))
        self.fields = all if self.fields is None else self.fields
        super().__init__(*args, **kwargs)

    def get_required_permissions(self, request):
        return [f"{self.model._meta.app_label}.change_{self.model.__name__.lower()}"]

    def get_context_data(self, *args, **kwargs):
        return {
            **super().get_context_data(*args, **kwargs),
            "layout": self._get_layout_cached(),
            "pagetitle": str(self.object),
        }

    def get_success_message(self, cleaned_data):
        return _("Saved %s" % self.object)

    # prevent browser caching edit views
    @never_cache
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)


def generate_copyview(model, attrs=None, labelfield=None, copy_related_fields=()):
    """
    creates a copy of a model instance and redirects to the edit view of the
    newly created instance
    attrs: custom field values for the new instance
    labelfield: name of a model field which will be used to create copy-labels
                (Example, Example (Copy 2), Example (Copy 3), etc)
    """
    attrs = attrs or {}

    def copy(request, pk: int):
        instance = get_object_or_404(model, pk=pk)
        if labelfield:
            attrs[labelfield] = copylabel(getattr(instance, labelfield))
        try:
            clone = deepcopy_object(instance, attrs, copy_related_fields)
        except Exception as e:
            messages.error(request, e)
            if request.GET.get("next"):
                return redirect(urllib.parse.unquote(request.GET["next"]))
            return redirect(reverse_model(model, "browse"))
        if request.GET.get("next"):
            return redirect(urllib.parse.unquote(request.GET["next"]))
        messages.success(request, _("Created copy %s" % instance))
        return redirect(reverse(model_urlname(model, "read"), args=[clone.pk]))

    return copy


def bulkcopy(request, queryset, attrs=None, labelfield=None, copy_related_fields=()):
    """creates a copy all instances of a queryset
    attrs: custom field values for the new instance
    labelfield: name of a model field which will be used to create copy-labels
                (Example, Example (Copy 2), Example (Copy 3), etc)
    copy_related_fields: related fields which should also be (deep) copied
    """
    attrs = attrs or {}
    created = 0
    errors = 0
    for instance in queryset:
        if labelfield:
            attrs[labelfield] = copylabel(getattr(instance, labelfield))
        try:
            deepcopy_object(instance, attrs, copy_related_fields)
            created += 1
        except Exception as e:
            messages.error(request, e)
            errors += 1
    messages.success(request, _("Created %s copies" % created))
    return redirect(request.path)


def deepcopy_object(instance, attrs=None, copy_related_fields=()):
    """
    Creates a 'deep' coyp of the model instance
    Related fields are only copied if they are listed in copy_related_fields
    ManyToMany relationships are copied automatically if they have no key in 'attrs'
    attrs defineds default values for fields
    """
    oldpk = instance.pk
    # see https://docs.djangoproject.com/en/3.2/topics/db/queries/#copying-model-instances
    many2manyfields = {
        f.name: getattr(instance, f.name).all()
        for f in instance._meta.get_fields()
        if f.many_to_many
    }
    instance.pk = None
    instance.id = None
    instance._state.adding = True
    for k, v in (attrs or {}).items():
        setattr(instance, k, v() if callable(v) else v)
    instance.save()
    for field, queryset in many2manyfields.items():
        getattr(instance, field).set(queryset)

    oldinstance = type(instance).objects.get(pk=oldpk)
    for field in copy_related_fields:
        related_name = oldinstance._meta.get_field(field).field.name
        for obj in getattr(oldinstance, field).all():
            obj.pk = None
            obj.id = None
            obj._state.adding = True
            setattr(obj, related_name, instance)
            obj.save()

    instance.save()
    return instance


def copylabel(original_name):
    """create names/labels with the sequence (Copy), (Copy 2), (Copy 3), etc."""
    copylabel = pgettext_lazy("this is a copy", "Copy")
    copy_re = f"\\({copylabel}( [0-9]*)?\\)"
    match = re.search(copy_re, original_name)
    if match is None:
        label = f"{original_name} ({copylabel})"
    elif match.groups()[0] is None:
        label = re.sub(copy_re, f"({copylabel} 2)", original_name)
    else:
        n = int(match.groups()[0].strip()) + 1
        label = re.sub(copy_re, f"({copylabel} {n})", original_name)
    return label



import random
import string
import traceback
import uuid

import django.urls
from django.contrib.admindocs.views import simplify_regex
from django.http import HttpResponseRedirect
from django.template import TemplateDoesNotExist
from django.test import Client, TestCase
from django_extensions.management.commands import show_urls

ROOT_URLPATTERNS = django.urls.get_resolver(None).url_patterns
ALPHANUMERIC_STR = "".join((string.ascii_letters, string.digits))


class TestAnonymousVisible(TestCase):
    # get a list of urlpatterns

    def test_visible(self):
        ALL_URLS = {
            x[2]: simplify_regex(x[1])  # { urlname: urlpattern }
            for x in show_urls.Command().extract_views_from_urlpatterns(
                ROOT_URLPATTERNS
            )
        }
        EXCEPTED_URLPATTERNS = {
            ALL_URLS[exclude]
            for exclude in (
                "admin:login",
                "login",
                "password_reset",
                "password_reset_done",
                "password_reset_complete",
                "password_reset_confirm",
            )
            if exclude in ALL_URLS
        }

        # update the exceptions for some pages that can be visible to the public
        # and those pages with dynamic url regex
        # and those out of the bread's scope (e.g., basxconnect)
        url_names = {ALL_URLS[x] for x in ALL_URLS} - EXCEPTED_URLPATTERNS

        # this tricky way works because '<' will always be encoded
        # if being requested, except for Django url patterns.
        url_with_args = {x for x in url_names if "<" in x}

        for url in url_with_args:
            args = [x for x in url.split("/")]
            for i in range(len(args)):
                if "<" not in args[i]:
                    continue

                if ":" in args[i]:  # the first one will specify its type
                    t = args[i][1 : args[i].find(":")]

                    if t == "int":
                        args[i] = str(random.randint(1, 100))  # nosec
                        # the line above can be avoided by bandit because
                        # it isn't for security purposes.
                    elif t == "uuid":
                        args[i] = str(uuid.uuid4())
                    else:
                        args[i] = "".join(random.sample(ALPHANUMERIC_STR, k=10))

                else:
                    args[i] = "".join(random.sample(ALPHANUMERIC_STR, k=10))

            url_names.add("/".join(args))

        # remove url with unsubstituted arguments
        url_names -= url_with_args

        # try fetching each url and see the response
        c = Client()

        for url in url_names:
            try:
                response = c.get(url)
                # if the status code is 4,
                # there's some errors loading those pages
                # However, 404 can happen with certain urls, make an exception
                if response.status_code == 404:
                    continue
                self.assertTrue(
                    not 400 <= response.status_code <= 499,
                    "This page %s failed to load, with status code %d."
                    % (url, response.status_code),
                )
                # if the status code is 3, and the Location is ALL_URLS["login"]
                # it's likely invisible to anonymous users, as the page
                # redirects to the login page.
                self.assertTrue(
                    300 <= response.status_code <= 399
                    and isinstance(response, HttpResponseRedirect)
                    and response.url.startswith(str(ALL_URLS["login"])),
                    "This page %s may be visible to anonymous users, with the status code %d"
                    % (url, response.status_code),
                )

            except TemplateDoesNotExist as e:
                traceback.print_exc()

                if hasattr(e, "message"):
                    print("The template for", url, "at", e.message, "does not exist.")
                else:
                    print("The template for", url, "does not exist.")



from django.utils.translation import gettext_lazy as _

from bread import menu, views
from bread.contrib.document_templates.models import DocumentTemplate
from bread.contrib.document_templates.views import (
    DocumentTemplateEditView,
    generate_document,
)
from bread.utils import Link, autopath, default_model_paths, model_urlname, urls
from bread.views.browse import delete
from bread.views.edit import bulkcopy, generate_copyview

urlpatterns = [
    *default_model_paths(
        DocumentTemplate,
        browseview=views.BrowseView._with(
            rowclickaction=views.BrowseView.gen_rowclickaction("edit"),
            bulkactions=[
                views.browse.BulkAction(
                    "delete", label=_("Delete"), iconname="trash-can", action=delete
                ),
                views.browse.BulkAction(
                    "copy",
                    label=_("Copy"),
                    iconname="copy",
                    action=lambda request, qs: bulkcopy(
                        request,
                        qs,
                        labelfield="name",
                        copy_related_fields=("variables",),
                    ),
                ),
            ],
        ),
        editview=DocumentTemplateEditView,
        readview=DocumentTemplateEditView,
        addview=views.AddView._with(fields=["name", "model", "file"]),
        copyview=generate_copyview(
            DocumentTemplate, labelfield="name", copy_related_fields=("variables",)
        ),
    ),
    autopath(generate_document, model_urlname(DocumentTemplate, "generate_document")),
]

menu.registeritem(
    menu.Item(
        Link(
            href=urls.reverse_model(DocumentTemplate, "browse"),
            label="Document Templates",
            iconname="document--blank",
        ),
        menu.Group("Document Templates", iconname="document--blank"),
    )
)
